"""Playyaml file.

version: 0.5.1
"""

import os
import sys
from pprint import pprint

import yaml

from . import config


def check_console():
    """Check if CONFIG.console is not none."""
    if config.CONSOLE is None:
        from rich.console import Console

        config.CONSOLE = Console()


def check_items_in_2_lists(list1, list2):
    """Check items in 2 lists returns True in at least one item matches.

    Args:
        list1: list of strings
        list2: list of strings

    Returns:
        bool: minimal one item matched
    """
    for i in list1:  # noqa:SIM110
        if i in list2:
            return True
    return False


def load_yaml(filename):
    """Load yaml file from disk.

    Args:
        filename: file to load

    Returns:
        struct from yaml file
    """
    try:
        with open(os.path.expanduser(filename), "r") as file:
            data = yaml.safe_load(file)
    except Exception as e:  # noqa:B902
        sys.stderr.write(f"can't open config file: {filename}  {e}")
        exit(1)
    return data


def check_arguments_in_task(task, command_dict):
    """Check arguments in task.

    Args:
        task: task array
        command_dict: command_dict array

    Returns:
        Boolean arguments correct
    """
    if "args" not in command_dict:
        # no arguments needed for this functions
        return True
    for arg in command_dict["args"]:
        argname = arg.get("name", "unknown_name")
        arghelp = arg.get("help", None)
        argdefault = arg.get("default", None)
        argtype = arg.get("type", "string")
        if argname in task:
            if argtype == "boolean":
                task[argname] = task[argname].lower() == "true"
            continue
        if argdefault is not None:
            task[argname] = argdefault
            if argtype == "boolean":
                task[argname] = task[argname].lower() == "true"
            continue
        # so if here we are missing a argname without a default so mandatoy
        sys.stderr.write(
            f"ERROR: missing argument {argname} for {arghelp or '<No Help>'} in {str(task)}"
        )
        return False
    return True


def task_check_tag(task, include_tags=None, exclude_tags=None):
    """Check if the task have a tag and if we skip or not this task.

    Args:
        task: task array
        include_tags: list of tags to run
        exclude_tags: list of tags to skip

    Returns:
        RunTask: boolean, True run task, False skip task
    """
    temp_tags = task.get("tags", [])
    tags = temp_tags if isinstance(temp_tags, list) else [temp_tags]

    if include_tags is not None and include_tags != []:  # noqa: SIM102
        # check is include tags in in tags of this task, if false then skip this task
        if check_items_in_2_lists(include_tags, tags) is False:
            return False

    if exclude_tags is not None and exclude_tags != []:  # noqa: SIM102
        # check is exclude tags in in tags of this task, if true then skip this task
        if check_items_in_2_lists(exclude_tags, tags) is True:
            return False

    return True


def loop_play_task(
    command_dict, task, *, include_tags=None, exclude_tags=None, funcdict
):
    """Handle a task that can loop.

    Args:
        command_dict: function info
        task: task array
        include_tags: list of tags to run
        exclude_tags: list of tags to skip
        funcdict: function dict

    Returns:
        Good: boolean
    """
    loop = command_dict.get("loop", None)
    if type(task[loop]) != list:
        print(f"loop variable {loop} in task but not a list.")
        exit(1)
    for i in task[loop]:
        new_task = task.copy()
        del new_task[loop]
        for p in i:
            new_task[p] = i[p]
        play_task(
            new_task,
            include_tags=include_tags,
            exclude_tags=exclude_tags,
            funcdict=funcdict,
        )
    return True


def play_loop(task):
    """Play command load_m3u_file.

    Args:
        task: task array

    Returns:
        Good: boolean
    """
    import copy

    with_items = task["with_items"]
    funcdict = task["funcdict"]
    for i in with_items:
        tasks = copy.deepcopy(task["tasks"])
        if type(i) == str:
            print(i)
            for loop_task in tasks:
                for key in loop_task:
                    if loop_task[key] == "__ITEM__":
                        loop_task[key] = i
                play_task(loop_task, funcdict=funcdict)
            pprint(tasks)
    return True


def play_task(task, *, include_tags=None, exclude_tags=None, funcdict):
    """Play a task.

    Args:
        task: task array
        include_tags: list of tags to run
        exclude_tags: list of tags to skip
        funcdict: function dict

    Returns:
        Good: boolean
    """
    check_console()
    task_name = task.get("name", task.get("command", "unknown"))
    if task_check_tag(task, include_tags=include_tags, exclude_tags=exclude_tags):
        print(f"running task: {task_name}")
    else:
        print(f"skipping task: {task_name}")
        return True

    if "command" not in task:
        sys.stderr.write(f"missing command entry in task {str(task)}")
        exit(3)

    if task["command"] == "loop":
        task["funcdict"] = funcdict

    command_dict = funcdict.get(task["command"], None)
    if command_dict is None:
        print("unknown task:  {task['command']}")
        pprint(task)
        return True

    loop = command_dict.get("loop", None)
    if loop is not None and loop in task:
        loop_play_task(
            command_dict,
            task,
            include_tags=include_tags,
            exclude_tags=exclude_tags,
            funcdict=funcdict,
        )
        return True

    if check_arguments_in_task(task, command_dict) is False:
        print("missing arguments in task.")
        exit(1)

    if "func" in command_dict:
        command_dict["func"](task)
    return True


def play(commandfile, *, include_tags=None, exclude_tags=None, funcdict):
    """Play commandfile.

    Args:
        commandfile: yaml file with instructions
        include_tags: list of tags to run
        exclude_tags: list of tags to skip
        funcdict: dicts with options

    Returns:
        None
    """
    check_console()
    data = load_yaml(commandfile)

    if "tasks" not in data:
        sys.stderr.write("missing tasks entrie in yaml file")
        exit(2)

    for task in data["tasks"]:
        play_task(
            task,
            include_tags=include_tags,
            exclude_tags=exclude_tags,
            funcdict=funcdict,
        )

    return None


def func_dict_parse(funcdict, for_type=None):
    """Parse func dict dict.

    Args:
        funcdict: dict with all the commands
        for_type: interactive or None at this moment.

    Returns:
        funcdict: parsed
    """
    new_funcdict = dict()
    for p in funcdict:
        cur_func = funcdict[p]
        if cur_func.get("hidden", False) is True:
            continue
        if (
            for_type == "interactive"
            and cur_func.get("interactive_hidden", False) is True
        ):
            continue
        new_funcdict[p] = cur_func
    new_funcdict["loop"] = {
        "args": [{"name": "with_items"}, {"name": "tasks"}],
        "func": play_loop,
        "help": "loop over multiple tasks",
    }
    return new_funcdict



"testtest""xmltv_channels script for fhs_xmltv_tools."""

from . import config  # noqa: F401
from .xmltv_programs import search_and_replace_channel_id
from xmltv.models.xmltv import DisplayName


def get_channel_by_channel_id(xmltv_data, channel_id):
    """Get channel from channel list with channel_id.

    Args:
        xmltv_data: data object with xmltv data.
        channel_id: channel id we want to search.

    Returns:
        channel
    """
    for my_channel in xmltv_data.channel:
        if my_channel.id == channel_id:
            return my_channel
    return None


def get_channel_index_by_channel_id(xmltv_data, channel_id):
    """Get channel index from channel list with channel_id.

    Args:
        xmltv_data: data object with xmltv data.
        channel_id: channel id we want to search.

    Returns:
        channel_index
    """
    for (i, my_channel) in enumerate(xmltv_data.channel):
        if my_channel.id == channel_id:
            return i
    return -1


def channels_remove_all_channels_not_in_list(xmltv_data, channel_set):
    """Remove all channels that are not in the list.

    Args:
        xmltv_data: data object with xmltv data.
        channel_set: set of channel id you want to keep

    Returns:
        xmltv_data
    """
    for p in range(len(xmltv_data.channel), 0, -1):
        if (xmltv_data.channel[p - 1]).id not in channel_set:
            del xmltv_data.channel[p - 1]
    return xmltv_data


def join_channels(xmltv_data, xmltv_data_add, only_channels=None):
    """Add the channels from xmltv_data_add to the channels from xmtv_data.

    Args:
        xmltv_data: data object with xmltv data.
        xmltv_data_add: data object with xmltv data jou want to add.

    Returns:
        xmltv_data
    """
    for p in xmltv_data_add.channel:
        if only_channels is not None:
            if p.id not in only_channels:
                continue
        xmltv_data.channel.append(p)
    return xmltv_data


def return_list(var_input):
    """Return a list a var as var is not a list.

    Args:
        var_input: variable input

    Returns:
        list
    """
    if type(var_input) != list():
        return [var_input]
    else:
        return var_input


def return_channel_displayname(p):
    """Return display info.

    Args:
        p: input from yaml

    Returns:
        displayname
    """
    if type(p) == str():
        return DisplayName(content=p, lang=None)
    if type(p) == dict():
        for q in p:
            return DisplayName(content=p[q], lang=q)
    return DisplayName(content=f"unknown {str(p)}")


def rename_channel(xmltv_data, channel):
    """Rename channel base on channel struct.

    Args:
        xmltv_data: data object with xmltv data.
        channel: channel struct

    Returns:
        xmltv_data
    """
    if type(channel) != dict:
        print('channel needs to by of type dict not {str(channel)}.')
        return xmltv_data

    for q in channel:
        nothing_found = True
        while (channel_index := get_channel_index_by_channel_id(xmltv_data, q)) != -1:
            nothing_found = False
            print(f"{q=}  {channel_index=}")
            if 'new_id' in channel[q]:
                search_and_replace_channel_id(xmltv_data, q, channel[q]['new_id'])
                xmltv_data.channel[channel_index].id = channel[q]['new_id']
            # check displayname
            if 'displayname' in channel[q]:
                displayname = return_list(channel[q]['displayname'])
                xmltv_data.channel[channel_index].display_name = []
                for p in displayname:
                    xmltv_data.channel[channel_index].display_name.append(return_channel_displayname(p))
            if channel[q].get('only_one', False) is True:
                break
        if nothing_found is True:
            print(f'channel with id {q}, not found, skipping')
        # check icon
        # check url
    return xmltv_data


def rename_channels(xmltv_data, channels):
    """Rename channels base on channels struct.

    Args:
        xmltv_data: data object with xmltv data.
        channels: channels struct

    Returns:
        xmltv_data
    """
    for channel in channels:
        xmltv_data = rename_channel(xmltv_data, channel)

    return xmltv_data



import textwrap
from jinja2 import Template
import sys
import cmd2
import os
import yaml
from pathlib import Path
from pprint import pprint

from .playyaml_lib import play_task
from .__version__ import __app_name__,__version__


def clean_task(task, orig_func):
    """Clean task.

    Args:
        task: dict with all the tasks.
        orig_func: default functions.

    Returns:
        cleaned taskdict
    """
    for arg in orig_func['args']:
        if arg['name'] in task:
            if task[arg['name']] == arg.get('default', '_____unknown_____'):
                del task[arg['name']]
    return task


def clean_tasks(taskdict, funcdict):
    """Clean tasks.

    Args:
        taskdict: dict with all the tasks.
        funcdict: default functions.

    Returns:
        cleaned taskdict
    """
    for t in taskdict:
        my_orig_func = funcdict[t['command']]
        clean_task(t, my_orig_func)
    return taskdict


def save_tasks(filename, taskdict, funcdict):
    """Save tasks to file.

    Args:
        filename: save filename
        taskdict: save dict to task
        funcdict: funcdict we are going to use to remove the default args from tasks
    """
    taskdict = clean_tasks(taskdict, funcdict)

    with Path(filename).open("w") as target:
        yaml.dump({'tasks': taskdict}, target)


def add_argument_from_dict(input):
    """Create arguments text from dict input.

    Args:
        input: dicts with arg info

    Returns:
        String
    """
    if 'help' in  input:
        help_str=f", help='{input['help']}'"
    else:
        help_str=""
    if 'default' in input:
        default_str=f", default='{input['default']}'"
    else:
        default_str=", required=True"
    return f"'--{input['name']}', type=str{default_str}{help_str}"


def create_cmd2_class_str(my_dict):
    test_str = textwrap.dedent("""\
    class CmdLineApp(cmd2.Cmd):
        def __init__(self, funcdict, hist_file, prompt):
            super().__init__()
            self.__funcdict = funcdict
            self.__tasks = []
            self.prompt = prompt
            cmd2.Cmd.__init__(self, persistent_history_file=hist_file, persistent_history_length=200)
            del cmd2.Cmd.do_edit, cmd2.Cmd.do_alias
            del cmd2.Cmd.do_macro, cmd2.Cmd.do_run_pyscript
            del cmd2.Cmd.do_shortcuts, cmd2.Cmd.do_run_script
        # Setting this true makes it run a shell command if a cmd2/cmd command doesn't exist
        # default_to_shell = True
        save_tasks_parser = cmd2.Cmd2ArgumentParser()
        save_tasks_parser.add_argument('--file', type=str, help='filename to save.', required=True)

        @cmd2.with_argparser(save_tasks_parser)
        def do_save_tasks(self, args):
            \"\"\"Save tasks file to disk, to reuse.\"\"\"
            save_tasks(args.file, self.__tasks, self.__funcdict)


        {% for key, my_func in my_dict.items() %}
        {% if 'args' in my_func -%}
        {{ key }}_parser = cmd2.Cmd2ArgumentParser()
        {% for item in my_func['args'] -%}
        {{ key }}_parser.add_argument({{ add_argument_from_dict(item) }})
        {% endfor -%}
        #
        @cmd2.with_argparser({{ key }}_parser)
        def do_{{ key }}(self, args):
            \"\"\"{{ my_func['help'] }}.\"\"\"
            task = {
                'command': '{{ key }}',
                {% for item in my_func['args'] -%}
                '{{ item['name'] }}': args.{{ item['name'] }},
                {% endfor -%}
            }
            self.__tasks.append(task)
            play_task(task, funcdict=self.__funcdict)
        {% else -%}
        # no args
        {% endif -%}
        {% endfor -%}

    """)
    j2_template = Template(test_str)
    func_dict = {
        'add_argument_from_dict': add_argument_from_dict
    }

    j2_template.globals.update(func_dict)


    return j2_template.render({'my_dict': my_dict})


def run_cmd2(funcdict):
    test_str = create_cmd2_class_str(funcdict)
    #print(test_str)
    hist_file = os.path.expanduser(f"~/.{__app_name__}.history")
    exec(test_str, None, globals())
    c = CmdLineApp(funcdict, hist_file, prompt=f"({__app_name__}) ") # cmdline
    sys.exit(c.cmdloop())



"""Functions to get info out of a xmltv."""

from pprint import pprint

from . import config  # noqa: F401


def get_info(input_var):
    """Return info from input.

    Args:
        input_var: program.title or program.desc

    Returns:
        info
    """
    if input_var == []:
        return ""
    try:
        if input_var[0].content == []:
            return ""
        return input_var[0].content[0]
    except KeyError:
        pprint(input_var)
    except IndexError:
        pprint(input_var)
    return "Error"


def get_program_title(program):
    """Return title from program.

    For now only the first but TODO: check lang

    Args:
        program: details of program

    Returns:
        program title
    """
    return get_info(program.title)


def get_program_desc(program):
    """Return desc from program.

    For now only the first but TODO: check lang

    Args:
        program: details of program

    Returns:
        program description
    """
    return get_info(program.desc)



"""Sqlalchemy module."""

from . import config

try:
    from sqlalchemy import create_engine
    from sqlalchemy.orm import sessionmaker
    from sqlalchemy.ext.declarative import declarative_base
    from sqlalchemy import Column, String
except ModuleNotFoundError:
    print("python: sqlalchemy library not found.")
    exit(1)


def create_engine_url(sqltype, sqlconnect):
    """Create a engine url.

    Args:
        sqltype: which type, for now: sqlite or sqlalchemy
        sqlconnect: how to connect, for sqlite, file path

    Returns:
        full sqlalchemy sqlite url
    """
    if sqltype == "sqlite":
        return f"sqlite:///{sqlconnect}"
    if sqltype == "sqlalchemy" or sqltype == "raw":
        return sqlconnect
    print(f"ERROR: unknown sqltype: {sqltype}, options are 'sqlite' or 'sqlalchemy'")
    exit(1)


def sql_create_engine(engine_url, echo=False):
    """Create sqlalchemy engine.

    Args:
        engine_url: sql engine url 'sqlite:///proef.db'
        echo: extra debug, default False

    Returns:
        engine or None if issue
    """
    engine = create_engine(engine_url, echo=echo)
    return engine


def sql_create_session(engine_url, engine=None):
    """Create sqlalchemy session.

    Args:
        engine_url: sql engine url
        engine: predefined engine or None

    Returns:
        session or None if issue
    """
    if engine is None:
        engine = sql_create_engine(engine_url)
    Session = sessionmaker(bind=engine)
    return Session()


def init_sql(engine_url):
    """Init sql.

    Args:
        engine_url: sql engine url

    Returns:
        session or None if issue
    """
    engine = sql_create_engine(engine_url, echo=False)
    if engine is None:
        return None
    config.SQL["engine"] = engine
    session = sql_create_session(None, engine=engine)
    if session is None:
        return None

    Base = declarative_base()

    class Program(Base):
        """Program sqlalchemy base."""

        __tablename__ = "programme"

        channel = Column(String, primary_key=True)
        start = Column(String, primary_key=True)
        stop = Column(String)
        title = Column(String)
        desc = Column(String)
        date = Column(String)

        def __repr__(self):
            """Representation for a program.

            Returns:
                string description
            """
            return f"Program {self.title} {self.start} {self.channel}"

    config.SQL["session"] = session
    Base.metadata.create_all(engine)
    config.SQL["program"] = Program

    return session



"""Xmltv to sql routines."""

from . import config


def save_xmltv_to_sql(xmltv_data, sqltype, sqlconnect):
    """Save xml to sql.

    Args:
        xmltv_data: data object with xmltv data.
        sqltype: type of connection for example sqlite
        sqlconnect: connect for sql, for example for sqlite, this is the db file inclusief path

    Returns:
        boolean saved
    """
    from .sql import init_sql, create_engine_url
    from .func_xmltv import get_program_desc, get_program_title

    sqlalchemy_url = create_engine_url(sqltype, sqlconnect)

    if init_sql(sqlalchemy_url) is None:
        print(f"ERROR: can't open sql url: {sqlalchemy_url}")
        return False
    Program = config.SQL["program"]
    # save xmltv data to sql
    for p in xmltv_data.programme:
        program = Program(
            channel=p.channel,
            start=p.start,
            stop=p.stop,
            title=get_program_title(p),
            desc=get_program_desc(p),
            date=p.date,
        )
        config.SQL["session"].merge(program)
    config.SQL["session"].commit()


def search_program_create_program_return(program_data):
    """Return search_program item.

    Args:
        program_data: item of xmltv_data.programma

    Returns:
        dicts with channel, title, discription, start, stop
    """
    ad = {
        "channel": program_data.channel,
        "title": program_data.title,
        "description": program_data.desc,
        "start": program_data.start,
        "stop": program_data.stop,
    }
    return ad


def search_programs_sql(sqltype, sqlconnect, search_str, force_case=False):
    """Search all the programs and return list of results.

    Args:
        sqltype: type of connection for example sqlite
        sqlconnect: connect for sql, for example for sqlite, this is the db file inclusief path
        search_str: string object or regex search.
        force_case: case sensitive search or not

    Returns:
        list of search_Program_create_program_return items
    """
    result = []
    from .sql import init_sql, create_engine_url

    sqlalchemy_url = create_engine_url(sqltype, sqlconnect)

    if init_sql(sqlalchemy_url) is None:
        print(f"ERROR: can't open sql url: {sqlalchemy_url}")
        exit(1)

    session = config.SQL["session"]
    Program = config.SQL["program"]

    if force_case:
        print("ERROR: force case not implmented.")
        exit(2)

    for p in session.query(Program).filter(
        (Program.title.like(f"%{search_str}%")) | (Program.desc.like(f"%{search_str}%"))
    ):
        result.append(search_program_create_program_return(p))
    return result


def delete_sql(sqltype, sqlconnect, days):
    """Delete sql entries older than days.

    Args:
        sqltype: type of connection for example sqlite
        sqlconnect: connect for sql, for example for sqlite, this is the db file inclusief path
        days: integer days to keep

    Returns:
        Result
    """
    from .sql import init_sql, create_engine_url
    from datetime import datetime, timedelta

    sqlalchemy_url = create_engine_url(sqltype, sqlconnect)

    if init_sql(sqlalchemy_url) is None:
        print(f"ERROR: can't open sql url: {sqlalchemy_url}")
        exit(1)

    session = config.SQL["session"]
    Program = config.SQL["program"]

    past_date = datetime.now() + timedelta(days=-days)
    past_date_str = past_date.strftime("%Y%m%d")

    print(f"cleaning (deleting) all entries before {past_date:%Y-%m-%d}")

    my_query = session.query(Program).filter(Program.start < past_date_str)
    my_query.delete()
    session.commit()
    return True



"""Console script for fhs_xmltv_tools."""
import pathlib

from xmltv import xmltv_helpers
from xmltv.models.xmltv import Tv
from .__version__ import __project_name__


def xmltv_load(xmltv_file):
    """List channels xml.

    Args:
        xmltv_file: file name to load xmltv data from

    Returns:
        xmltv data array

    """
    xmltv_in_file = pathlib.Path(xmltv_file)
    data = xmltv_helpers.serialize_xml_from_file(xmltv_in_file, Tv)
    return data


def xmltv_save(xmltv_file, xmltv_data):
    """Write channels xml.

    Args:
        xmltv_file: file name to write xmltv data to
        xmltv_data: data object with xmltv data.
    """
    xmltv_out = pathlib.Path(xmltv_file)
    xmltv_helpers.write_file_from_xml(xmltv_out, xmltv_data)


def xmltv_empty(source_data_url="", source_info_name="", source_info_url=""):
    """Create empty xmltv forum.

    Args:
        source_data_url: url of source
        source_info_name: source name
        source_info_url: url of your source

    Returns:
        xmltv source
    """
    empty_xmltv = Tv()
    empty_xmltv.source_data_url = source_data_url
    empty_xmltv.source_info_name = source_info_name
    empty_xmltv.source_info_url = source_info_url
    empty_xmltv.generator_info_url = "https://github.com/foxhunt72/fhs-xmltv-tools"
    empty_xmltv.generator_info_name = __project_name__
    return empty_xmltv
